%{ 
 ****************************************************************************
 * CLASSIFICATION: Unclassified Unlimited Release
 *
 * SAND2020-0528 O
 *
 ****************************************************************************
%}

%  parseSphFileName: Parse different SPH file name formats and extract as
%  much information as possible. Returns defaults for values it's couldn't
%  extract. 
% 
%  sphFile = parseSphFileName(sphFileName)
%
%  Inputs:
%     sphFileName: requires .sph extension on name ex: '0806F09_PS0003_CC000000_N04_M3.sph'
%
%  Outputs (with default values):
%     sphFile.mission = 'MISSION'
%     sphFile.passName = 'PS0000'
%     sphFile.passNum = 0
%     sphFile.patchNum = 0
%     sphFile.continuousCollectNum = 0
%     sphFile.nodeNum = 3             
%     sphFile.moduleNum = 1           
%     sphFile.channelNum = -1       
%     sphFile.continuousCollectBaseName = ''
%     sphFile.matchFound = 1;
%     sphFile.isContinuousCollect = 0

function sphFile = parseSphFileName(sphFileName)
    sphFile.mission = 'MISSION';
    sphFile.passName = 'PS0000';
    sphFile.passNum = 0;
    sphFile.patchNum = 0;
    sphFile.continuousCollectNum = 0;       % Only set for multi-part pass files
    sphFile.nodeNum = 3;                    % Default single processing node
    sphFile.moduleNum = 1;                  % Default single module
    sphFile.channelNum = -1;                % -1 to be clear that channelNum wasn't part of filename
    sphFile.continuousCollectBaseName = ''; % Base name that can be used to search for all CC files from the same pass
    sphFile.matchFound = 1;
    sphFile.isContinuousCollect = 0;
    
    [~, name, ~] = fileparts(sphFileName);
          
    %%
    % 20130305.0305KUa.PASS1015.00362
    % 20110226.A02NM01.PASS0041.00000_0
    tok = regexp(name, '^(?<date>[^.]+)\.(?<mission>[^.]+)\.PASS(?<passNum>[^.]+)\.(?<patchNum>[^.]+)$', 'names');
    if ~isempty(tok)
        sphFile.mission = tok.mission;       
        sphFile.passName = ['PASS' tok.passNum];
        sphFile.passNum = str2double(tok.passNum);

        % Patch number will be '00323' format if patch mode,
        % '00000_0' if continuous collect. Determine which it is.
        patchTok = regexp(tok.patchNum, '[^_]+', 'match');
        if size(patchTok, 2) == 1
            % Simple patch number, just return it
            sphFile.patchNum = str2double(tok.patchNum);
        else
            % Continuous collect patch number '00000_0' or '00233_1'
            % Only return file item for first file in continuous collect pass
            sphFile.patchNum = 0;
            sphFile.continuousCollectNum = str2double(patchTok{2});
            sphFile.isContinuousCollect = 1;
            
            % Returns 20110226.A02NM01.PASS0041 from 20110226.A02NM01.PASS0041.00000_0 
            baseTok = regexp(name, '^(?<baseName>.*)\..*$', 'names');
            sphFile.continuousCollectBaseName = baseTok.baseName;
        end
        
        return;
    end        
                        
    %%
    % *Patch00001
    % *Patch00000_0, *Patch00021_1
    tok = regexp(name, '.*Patch(?<patchStr>.*)$', 'names');
    if ~isempty(tok)
        % Check whether this is normal patch or continuous collect
        splitStr = strsplit(tok.patchStr, '_');
                
        if length(splitStr) == 1
            % No _ was found, so simple patch
            sphFile.patchNum = str2double(splitStr{1}); 
            sphFile.continuousCollectNum = 0;
        else
            % Continuous collect it is
            sphFile.patchNum = 0;
            sphFile.continuousCollectNum = str2double(splitStr{2});
            sphFile.isContinuousCollect = 1;
                                    
            % Returns *Patch00000 from *Patch00000_0
            baseTok = regexp(name, '(?<baseName>.*Patch)\d*\_.*$', 'names');
            sphFile.continuousCollectBaseName = baseTok.baseName;
        end
        
        return;
    end
    
    %%
    % 0502MRa_PS0001_CC000000
    % 0502MRa_PS0001_PT000001
    % 0502MRa_PS0001_PT000001_CH0
    tok = regexp(name, '^(?<mission>.+)\_PS(?<passNum>[^_]+)\_(?<patchStr>\D\D)(?<patchNum>[^_]+)$', 'names');
    if ~isempty(tok)
        sphFile.mission = tok.mission;
        sphFile.passName = ['PS' tok.passNum];
        sphFile.passNum = str2double(tok.passNum);
        
        tok.patchNum = str2double(tok.patchNum);
        
        % Interpret the patch to determine if it's patch or CC
        if strcmp(tok.patchStr, 'PT')            
            sphFile.patchNum = tok.patchNum;
            sphFile.continuousCollectNum = 0;
        elseif strcmp(tok.patchStr, 'CC')
            sphFile.patchNum = 0;
            sphFile.continuousCollectNum = tok.patchNum;            
            sphFile.isContinuousCollect = 1;
            
            % Returns 0502MRa_PS0001_CC from 0502MRa_PS0001_CC000000
            baseTok = regexp(name, '(?<baseName>.*CC)\d*$', 'names');
            sphFile.continuousCollectBaseName = baseTok.baseName;
        end
        
        return;
    end
    
    %% Not sure if this has ever been a valid filename
    % 0502MRa_PS0001_PT000001_CH0
    tok = regexp(name, '^(?<mission>.+)\_PS(?<passNum>[^_]+)\_(?<patchStr>\D\D)(?<patchNum>[^_]+)\_CH(?<channelNum>.*)$', 'names');
    if ~isempty(tok)
        sphFile.mission = tok.mission;
        sphFile.passName = ['PS' tok.passNum];
        sphFile.passNum = str2double(tok.passNum);
        
        tok.patchNum = str2double(tok.patchNum);
        
        sphFile.channelNum = str2double(tok.channelNum);
        
        % Interpret the patch to determine if it's patch or CC
        if strcmp(tok.patchStr, 'PT')            
            sphFile.patchNum = tok.patchNum;
            sphFile.continuousCollectNum = 0;
        elseif strcmp(tok.patchStr, 'CC')
            sphFile.patchNum = 0;
            sphFile.continuousCollectNum = tok.patchNum;            
            sphFile.isContinuousCollect = 1;
            
            % Returns 0502MRa_PS0001_CC from 0502MRa_PS0001_CC000000
            baseTok = regexp(name, '(?<baseName>.*CC)\d*$', 'names');
            sphFile.continuousCollectBaseName = baseTok.baseName;
        end
        
        return;
    end
            
    %%
    % 0502MRa_PS0001_CC000000_N03_M1
    % 0502MRa_PS0001_PT000001_N03_M1
    tok = regexp(name, '^(?<mission>.+)\_PS(?<passNum>[^_]+)\_(?<patchStr>\D\D)(?<patchNum>[^_]+)\_N(?<nodeNum>[^_]+)\_M(?<moduleNum>[^_]+)$', 'names');
    if ~isempty(tok)
        sphFile.mission = tok.mission;
        sphFile.passName = ['PS' tok.passNum];
        sphFile.passNum = str2double(tok.passNum);
        sphFile.nodeNum = str2double(tok.nodeNum);
        sphFile.moduleNum = str2double(tok.moduleNum);
        
        tok.patchNum = str2double(tok.patchNum);
        
        % Interpret the patch to determine if it's patch or CC
        if strcmp(tok.patchStr, 'PT')            
            sphFile.patchNum = tok.patchNum;
            sphFile.continuousCollectNum = 0;
        elseif strcmp(tok.patchStr, 'CC')
            sphFile.continuousCollectNum = tok.patchNum;
            sphFile.patchNum = 0;
            sphFile.isContinuousCollect = 1;
            
            % Returns 0502MRa_PS0001_CC from name 
            baseTok = regexp(name, '(?<baseName>.*CC).*$', 'names');
            sphFile.continuousCollectBaseName = baseTok.baseName;
        end
        
        return;
    end        
    
    %%
    % Continuous patch collect
    % 1112MRz_PS0148_PT000001_PART000_N03_M1
    % 1112MRz_PS0148_PT000001_PART001_N03_M1
    tok = regexp(name, '^(?<mission>.+)\_PS(?<passNum>[^_]+)\_PT(?<patchNum>[^_]+)\_PART(?<partNum>[^_]+)\_N(?<nodeNum>[^_]+)\_M(?<moduleNum>[^_]+)$', 'names');
    if ~isempty(tok)
        sphFile.mission = tok.mission;
        sphFile.passName = ['PS' tok.passNum];
        sphFile.passNum = str2double(tok.passNum);
        sphFile.nodeNum = str2double(tok.nodeNum);
        sphFile.moduleNum = str2double(tok.moduleNum);
                
        % Multipart files are continuous collect, however they have
        % both a patch and continuousCollectNum
        sphFile.patchNum = str2double(tok.patchNum);
        sphFile.continuousCollectNum = str2double(tok.partNum);
        sphFile.isContinuousCollect = 1;

        % Returns 1112MRz_PS0148_PT000001_PART from name
        baseTok = regexp(name, '(?<baseName>.*PART).*$', 'names');
        sphFile.continuousCollectBaseName = baseTok.baseName;
       
        return;
    end
    
    display(sprintf('File %s did not match any known SPH file name filters, all values set to defaults.', sphFileName));
    sphFile.matchFound = 0;
end
