%{ 
 ****************************************************************************
 * CLASSIFICATION: Unclassified Unlimited Release
 *
 * SAND2020-0528 O
 *
 ****************************************************************************
%}

% READ_SPH_STREAM Class to write SPH files according to format defined in the preamble
%
% Given a Sandia SPH file FILENAME, this function reads the preamble of the
% file and the phase history and related binary data contained from pulse
% FIRST_PULSE to (FIRST_PULSE+PULSE_COUNT-1).
%
% SPH files are the preferred format for distribution of Sandia National
% Laboratories radar phase history. SPH files consist of a ASCII preamble
% followed by binary data.
%
% The format is described in detail in:
%    "SPH Specifications Version 1.1"
%
% As SPH is more or less a self-documenting format this reader should
% require minimal or no change when new fields are added to SPH. As it
% builds the parse from the SPH file itself it may be a bit slower than the
% others, and the resulting structure may be more 'fluid' between files and
% revisions.
%
% When modifying this file please note that it is important to use
% *integers* and not doubles for finding byte offsets in large files.
% Otherwise round off error will cause serious problems!
%
% KNOWN BUGS:
%     BitDefs are not fully supported
%
%     Error checking is currently incomplete
%
% Author: Rick Naething
%
%  Date      Revision History
%---------   -------------------------------------------------------------------
% 06/29/09   rmn - Created
% 11/03/11   rmn - Added support for UserDef / Arbitrary Structs
% 06/28/12   rmn - Completed Support for Multipart files
% 01/30/15   jwe - Modified to use sph.m

classdef read_sph_stream < sph
   
    properties
        filenames
        debug
        file_reader
        total_pulses
        curr_pulse

        % User Accessible Data
        preamble
        Const
        Data
    end
    
    properties(Access = private)
        % file information
        sph_fid
        curr_file
        machine_format
        byte_lookup
        sample_data_start_byte
        pulse_reader;
                
        % general use
        sph_types 
        pulse_byte_length = uint64(0);
    end
        
    methods

        %---------------------------------------------------------------------%
        % constructor
        function this = read_sph_stream(filename, readHeaderOption)
            
            %----------------%
            % Initialize
            this.debug                  = 0;
            this.file_reader            = 0;
            this.total_pulses           = uint64(0);
            this.curr_pulse             = uint64(1); % Matlab indexing, starting from 1!
            this.sph_fid                = 0;
            this.curr_file              = 1;         % Matlab indexing, starting from 1!
            this.machine_format         = 'native';
            this.sample_data_start_byte = uint64(0);
            this.pulse_byte_length      = uint64(0);
            
            if ~exist('filename','var')
                [name path] = uigetfile('*.sph','Please select the debug file: '); 
                if ( name == 0 )
                    error('File selection canceled by user.');        
                end
                filename = fullfile(path, name);
            end
            
            % This is used to skip past complex data
            if ~exist('readHeaderOption','var')
                readHeaderOption = 'all';
            end
            
            %----------------%
            % Handle single / multi-part SPH files
            
            [sph_dir, file, ext] = fileparts(filename);
            
            % This handles all the different SPH name formats and returns the necessary info
            sphFile = parseSphFileName(filename);
            
            % Check if this is a multipart file, and if so read in all the parts
            if sphFile.isContinuousCollect
                allFiles = dir(fullfile(sph_dir, [sphFile.continuousCollectBaseName '*' ext]));
                
                % Make sure the file exists
                if length(allFiles) == 0
                    fprintf('\nERROR Could not find file %s\n', filename);
                    return;
                end
                
                % Read in all of the multipart files found above
                vectorIdx = 1;
                for idx = 1:length(allFiles)
                    sphFileTmp = parseSphFileName(allFiles(idx).name);
                    if (sphFileTmp.nodeNum == sphFile.nodeNum) && (sphFileTmp.moduleNum == sphFile.moduleNum)
                        fprintf('Adding file %s\n', allFiles(idx).name);
                        this.filenames{vectorIdx} = fullfile(sph_dir, allFiles(idx).name);
                        vectorIdx = vectorIdx + 1;
                    else
                        fprintf('Skipping file %s as it came from a different node or module\n', allFiles(idx).name);
                    end
                end                
            else
                % Otherwise it's just the one file
                this.filenames = {filename};
            end

            this.build_byte_lookup();

            %----------------%
            % init file, read preamble
            this.open_file(1);
            this.preamble = sph.read_sph_preamble(this.sph_fid);
            
            % read in binary info
            this.sph_types = sph.sph_data_types(); % Types are defined externally
            this.pulse_byte_length = sum([this.preamble.datastructs.PulseStruct.bytes]);

            % Read from the beginning of the binary data
            this.determine_endianess();
            this.read_constants();
            
            this.total_pulses = (this.byte_lookup(end).end_bytes - ftell(this.sph_fid)) / this.pulse_byte_length;
            
            this.sample_data_start_byte = ftell(this.sph_fid);
            
            % Synthesize reader function, tell it whether to skip the complex data or not
            this.pulse_reader = this.generate_pulse_reader(this.preamble, read_sph_stream.getHeaderFlag(readHeaderOption, this.preamble.Mode));
        end
        
        %---------------------------------------------------------------------%
        function close(this)
            this.close_file();
        end
        
        %---------------------------------------------------------------------%
        function generate_file_reader(this)
            warning('File based pulse reader is no longer needed for fast operation, just debugging');
            fid = fopen('temp_pulse_reader.m', 'w');
            fprintf(fid, 'function Data = temp_pulse_reader(sph_fid, Data, offset, pulse_count, n_chan)\n\n');
            fprintf(fid, regexprep(this.pulse_reader, 'this.Data','Data'));
            fprintf(fid, '\nend\n');
            fclose(fid);
            rehash;
            this.file_reader = 1;
        end
        
        %---------------------------------------------------------------------%
        function goto_pulse(this, pulse)
            this.curr_pulse = uint64(pulse);
            byte_offset = this.sample_data_start_byte + (this.curr_pulse-1)*this.pulse_byte_length;
            this.goto_byte(byte_offset);
        end
        
        %---------------------------------------------------------------------%
        function goto_file(this, file_num)
            this.close_file();
            this.open_file(file_num);
        end        
                
        %---------------------------------------------------------------------%
        function read_pulses(this, n_pulses)
            
            % If not specified read all pulses
            if ~exist('n_pulses','var')                
                n_pulses = this.total_pulses;
                display(sprintf('Number of pulses not specified, reading all %d pulses', n_pulses));
            end        
            
            total_pulses_left = (this.total_pulses - this.curr_pulse + 1);
            if (n_pulses > total_pulses_left)
                n_pulses = total_pulses_left;
                fprintf(2, 'Too many pulses requested, reading all remaining pulses in file (%d)\n', n_pulses);
            end
            
            this.Data = [];
            
            % init ph storage
           % this.Data.SampleData = ...
           %     complex(zeros(this.preamble.Nsamples, n_pulses, this.preamble.Nchannels));

            % Read pulses from file
            pulses_needed    = uint64(n_pulses);
            start_offset     = 1;
            
            while(pulses_needed)           
                % Determine how many to read from current file
                pulses_available = (this.byte_lookup(this.curr_file).bytes - ftell(this.sph_fid)) ./ this.pulse_byte_length;
                pulsesToRead     = min(pulses_needed, pulses_available);

                % Return if the user asked for more pulses than are available
                if(pulsesToRead==0)
                    break;
                end

                % Read pulses from file
                disp(['Reading ' num2str(pulsesToRead) ' pulses']);       
                this.read_pulse_data(start_offset, pulsesToRead);
                pulses_needed   = pulses_needed - pulsesToRead;
                start_offset    = start_offset + pulsesToRead;

                % Advance to next file if needed
                this.goto_pulse(this.curr_pulse + pulsesToRead);
            end

        end
            
    end
    
    methods (Access = private)
        
        %---------------------------------------------------------------------%
        function read_constants(this)
            if isfield(this.preamble.datastructs, 'ConstStruct')
                for c = 1:length(this.preamble.datastructs.ConstStruct)
                    this.Const.(this.preamble.datastructs.ConstStruct(c).name) = ...
                        this.readdata(this.preamble.datastructs.ConstStruct(c));
                end
            end
        end
        
        %---------------------------------------------------------------------%        
        function open_file(this, file_index)
            this.curr_file = file_index;
            file = this.filenames{file_index};
            try
                this.sph_fid = fopen(file,'r', this.machine_format);
                disp(['Opened: ' file]);
            catch ME1
                disp('Unable to open SPH File');
                rethrow(ME1);
            end
        end

        %---------------------------------------------------------------------%
        function close_file(this)
            try
                fclose(this.sph_fid);
            catch ME1
                disp('Unable to close SPH File');
                rethrow(ME1);
            end
        end
    
        %---------------------------------------------------------------------%
        function build_byte_lookup(this)
            pre_bytes = uint64(0);
            for f=1:length(this.filenames)
                d = dir(this.filenames{f});
                if(~size(d,1))
                    error(['Could not open file: ' this.filenames{f}]);
                end
                this.byte_lookup(f).beg_bytes = pre_bytes;
                pre_bytes = pre_bytes + uint64(d.bytes);
                this.byte_lookup(f).end_bytes = pre_bytes;
                this.byte_lookup(f).bytes     = uint64(d.bytes);
            end
        end       

        %---------------------------------------------------------------------%
        function read_pulse_data(this, offset, pulse_count)

            if(this.file_reader)
                this.Data = temp_pulse_reader(this.sph_fid, this.Data, offset, pulse_count, this.preamble.Nchannels);
                return;
            else
                sph_fid = this.sph_fid;           %#ok; needed for eval
                n_chan = this.preamble.Nchannels; %#ok; needed for eval
                eval(this.pulse_reader);
            end

        end

        %---------------------------------------------------------------------%
        function determine_endianess(this) 
            magic = fread(this.sph_fid, 4, '*char').';
            switch(magic)
                case 'ojoM'
                    mf='ieee-le';
                case 'Mojo'
                    mf='ieee-be';
                otherwise
                    disp('Error Reading File');
                    return;
            end
            this.preamble.dataFormat = mf;
            this.machine_format = mf;
            pos = ftell(this.sph_fid);
            this.close_file();
            
            % reopen with currect endianess
            this.open_file(1);
            fseek(this.sph_fid, pos, 'bof');
        end
        
        %---------------------------------------------------------------------%
        function t_f = byte_in_file(this, byte, f_idx)
            t_f = ((this.byte_lookup(f_idx).beg_bytes <= byte) && ...
                   (this.byte_lookup(f_idx).end_bytes >  byte));
        end
        
        %---------------------------------------------------------------------%
        function goto_byte(this, byte)

            % If this byte is in the current file
            if ( this.byte_in_file(byte, this.curr_file) )
                % no-op
            else
                for f=1:length(this.byte_lookup)
                   if (this.byte_in_file(byte, f))
                       this.close_file();
                       this.open_file(f);
                   end
                end
            end

            try
                byte = byte - this.byte_lookup(this.curr_file).beg_bytes;
                fseek(this.sph_fid, uint64(byte), 'bof');
            catch ME1
                disp('Error seeking in SPH file');
                rethrow(ME1);
            end
        end

        %---------------------------------------------------------------------%
        function data=readdata(this, chunk)
            switch(chunk.type)
                
                % these types agree between matlab and the SPH format
                case fieldnames(this.sph_types.real)
                    data=fread(this.sph_fid, chunk.bytes/this.sph_types.real.(chunk.type), chunk.type);

                case fieldnames(this.sph_types.cmpl)
                    data = this.read_complex(2*chunk.bytes/this.sph_types.cmpl.(chunk.type).size,...
                                        this.sph_types.cmpl.(chunk.type).type);

                case {'SPHPos'}
                    data.Status = fread( this.sph_fid, 1, 'int8');
                    data.xat    = fread( this.sph_fid, 1, 'float64');
                    data.yon    = fread( this.sph_fid, 1, 'float64');
                    data.zae    = fread( this.sph_fid, 1, 'float64');

                otherwise
                    if (isfield(this.preamble.datastructs,chunk.type))
                        for idx = 1:length(this.preamble.datastructs.(chunk.type)),
                            n = this.preamble.datastructs.(chunk.type)(idx).name;
                            t = this.preamble.datastructs.(chunk.type)(idx);
                            data.(n) = this.readdata(t);
                        end
                    else
                        warning('ERROR: Procecessing of %s not yet implemented', chunk.type);
                    end

            end
        end

        %---------------------------------------------------------------------%
        function data = read_complex(this, length, type)
            comp=fread(this.sph_fid, length, type);
            data=complex(comp(1:2:end),comp(2:2:end));
        end        
    end
    
    methods(Static)
        
        %---------------------------------------------------------------------%
        function headerFlag = getHeaderFlag(headerString, modeString)
                                    
            if strcmp(headerString, 'all')
                % Complex data (default)
                headerFlag = 1;            
            elseif strcmp(headerString, 'none')
                % No complex data
                headerFlag = 0;            
            elseif strcmp(headerString, 'patchOnly')
                % Only load complex data if it's not a continuous collect mode
                
                % Convert preamble radar mode to enum and check if it's continuous collect
                if radarModeIsContinuousCollect(radarMode2Enum(modeString))
                    % No complex for continuous collect
                    headerFlag = 0;
                else
                    headerFlag = 1;
                end
            else
                warning('ERROR: headerString %s not valid. Valid values are [none, all, patchOnly]. Defaulting to all', ...
                        headerString);
                headerFlag = 1;
            end            
        end
        

        %---------------------------------------------------------------------%
        %---------------------------------------------------------------------%
        % Pulse Reader Generation
        %---------------------------------------------------------------------%
        %---------------------------------------------------------------------%
        function buffer = generate_pulse_reader(preamble, includeHeaders)
            
            % This is used to skip past complex data
            if ~exist('includeHeaders','var')
                includeHeaders = 1; 
            end;
            
            buffer = '';
            start_ = 1;
            
            bytes = preamble.datastructs.PulseStruct.bytes / preamble.Nchannels;
            bprintf('    buff = fread(sph_fid, [%d, n_chan*pulse_count], ''*uint8'');\n\n', bytes);
            bprintf('    for chan = 1:n_chan\n');
            
            PulseStruct = preamble.datastructs.PulseStruct;
            ChanStruct  = preamble.datastructs.ChanStruct;
            sph_types = sph.sph_data_types(); %#ok;
            
            for datachunk = 1:length(PulseStruct)
                
                switch PulseStruct(datachunk).type
                    case 'ChanStruct'
                        for j = 1:length(ChanStruct)
                            switch ChanStruct(j).name
                                case 'SampleData'
                                    sidx   = sprintf('(1:%d, offset+(0:pulse_count-1), chan)', preamble.Nsamples);
                                    read_data('this.Data.SampleData',sidx, ChanStruct(j));
                                otherwise
                                    count = find_count(ChanStruct(j));
                                    lhand = sprintf('this.Data.%s', ChanStruct(j).name);
                                    sidx   = sprintf('(1:%d, offset+(0:pulse_count-1), chan)', count);
                                    read_data(lhand,sidx,ChanStruct(j));
                            end
                        end
                        
                    case sph_types.all
                        switch PulseStruct(datachunk).name
                            case 'SampleData'
                                read_data('this.Data.SampleData','(:,pulse,1)',ChanStruct(1));
                            otherwise
                                lhand = sprintf('this.Data.%s',PulseStruct(datachunk).name);
                                sidx  = '(pulse)';
                                read_data(lhand, sidx, PulseStruct(datachunk));
                        end
                    otherwise
                        warning(1, ['ERROR: Support for ' PulseStruct(datachunk).type ' not yet added']);
                end
            end
            bprintf('    end\n');
            
            %---------------------------------------------------------------------%
            function c=find_count(s)
                switch s.type
                    case fieldnames(sph_types.real)
                        sz = sizeof(s.type);
                    case fieldnames(sph_types.cmpl)
                        sz = sizeof(sph_types.cmpl.(s.type).type)./2;
                    case 'SPHPos'
                        sz = 25;
                    otherwise
                        sz = sum([preamble.datastructs.(s.type).bytes]);
                end
                c = s.bytes/sz;
            end
            
            %---------------------------------------------------------------------%
            function s=sizeof(type)
                type = conv_type(type);
                t = typecast([], type);
                t(1)=0; %#ok;
                t=whos('t');
                s=t.bytes;
            end
            
            %---------------------------------------------------------------------%
            function type=conv_type(type)
                type = regexprep(type, 'float32', 'single');
                type = regexprep(type, 'float64', 'double');
            end
            
            %---------------------------------------------------------------------%
            function bprintf(formatSpec, varargin)
                buffer = [buffer sprintf(formatSpec, varargin{:})];
            end
            
            %---------------------------------------------------------------------%
            function read_data(lhand, idx_str, chunk)
                                
                switch(chunk.type)
                    
                    % these types agree between matlab and the SPH format
                    case fieldnames(sph_types.real)
                        pp_assign(strcat(lhand, idx_str), read(chunk.bytes/sph_types.real.(chunk.type), chunk.type, chunk.bytes));
                        
                    case fieldnames(sph_types.cmpl)
                        l = 2*chunk.bytes/sph_types.cmpl.(chunk.type).size;
                        t = sph_types.cmpl.(chunk.type).type;
                        
                        % If we only want the headers advance the file position but don't assign the data                        
                        if (includeHeaders)
                            pp_assign('comp', read(l, t, sizeof(t)));
                            rhs = sprintf('reshape(complex(double(comp(1:2:end)),double(comp(2:2:end))), [%d pulse_count])', preamble.Nsamples);
                            pp_assign(strcat(lhand, idx_str), rhs);
                        else
                            % Don't use the output of the read command, just let it adjust the start/end pointers
                            % in the data so the next read is from the correct position
                            read(l, t, sizeof(t));
                        end
                        
                    case {'SPHPos'}
                        pp_assign(strcat(lhand, '.Status', idx_str), read(1, 'int8',    1));
                        pp_assign(strcat(lhand, '.xat',    idx_str), read(1, 'float64', 8));
                        pp_assign(strcat(lhand, '.yon',    idx_str), read(1, 'float64', 8));
                        pp_assign(strcat(lhand, '.zae',    idx_str), read(1, 'float64', 8));
                        
                    otherwise
                        if (isfield(preamble.datastructs,chunk.type))
                            for idx = 1:length(preamble.datastructs.(chunk.type)),
                                n = preamble.datastructs.(chunk.type)(idx).name;
                                t = preamble.datastructs.(chunk.type)(idx);
                                read_data(strcat(lhand, '.', n), idx_str, t);
                            end
                        else
                            warn(['ERROR: Procecessing of ' chunk.type 'not yet implemented']);
                        end
                        
                end
                
                %---------------------------------------------------------------------%
                function pp_assign(lhand, rhand)
                    bprintf('        %-60s = %-50s\n', lhand, strcat(rhand,';'));
                end
                
                %---------------------------------------------------------------------%
                function str = read(count, type, bytes)                   
                    type=conv_type(type);
                    end_ = start_ + (count*bytes);
                    str = sprintf('typecast(reshape(buff(%d:%d, chan:n_chan:end), [1 %d*pulse_count]), ''%s'')', start_, end_-1, count*bytes, type);
                    start_ = end_;
                end
                
            end    
        
        end
    end
end






