%{ 
 ****************************************************************************
 * CLASSIFICATION: Unclassified Unlimited Release
 *
 * SAND2020-0528 O
 *
 ****************************************************************************
%}


%This file is used internally by the sph reader and writer utilities
%
% Author: Rick Naething
%
%  Date      Revision History
%---------   -------------------------------------------------------------------
% 02/08/12   rmn - Created
% 04/27/16   rmr - changed 'cmplxf' type from 'float32' to 'single' and
%                  'cmplxd' type from 'float64' to 'double'

classdef sph < handle
    
    methods(Static, Access=public)
    
        %=================================================================%
        function types = sph_data_types()

            types.real = struct('uint8',1,...
                                  'uint16',2,...
                                  'uint32',4,...
                                  'uint64',8,...
                                  'int8',1,...
                                  'int16',2,...
                                  'int32',4,...
                                  'int64',8,...
                                  'float32',4,...
                                  'float64',8);

            types.cmpl = struct('cmplxs', struct('size', 4,  'type', 'int16'),...
                                'cmplxi', struct('size', 8,  'type', 'int32'),...
                                'cmplxf', struct('size', 8,  'type', 'single'),...
                                'cmplxd', struct('size', 16, 'type', 'double'));

            types.union = struct('SPHPos', 25);

            types.all   = [fieldnames(types.real);
                           fieldnames(types.cmpl);
                           fieldnames(types.union)];
        end
        
        %=================================================================%
        function tokens = sph_tokens()

            % tokens that contain a number
            tokens.number={'StartTime';
                           'LightSpeed';
                           'Npulses';
                           'Nsamples';
                           'Nchannels';
                           'NumPulsesInAir';
                           'DesCntrFreq';
                           'DesGTA';
                           'DesElevAng';
                           'DesRange';
                           'DesGrndSqnt';
                           'DesAzRate';
                           'DesAzRes';
                           'DesRngRes';
                           'MagCal';
                           'SigmaN';
                           'ADF';
                           'SystemSerialNum'};
                       
            % tokens that contain a string
            tokens.word={'Mission';
                         'Sensor';
                         'SoftwareVersionNum'};

            % tokens that contain a data type
            tokens.gpstype={'SRP'};

            % Tokens that define types
            tokens.data_type = {'ChirpRate';
                                'ChirpRateDelta';
                                'StartF';
                                'StartFDelta';
                                'RxPos';
                                'TxPos';
                                'TRDelay';
                                'RxTime';
                                'RxTimeDelta';
                                'TxTime';
                                'SampleData';
                                'ChannelData';
                                'SRP';
                                'AntPnt';
	                            'DesValidPulseMotionInterval';
	                            %% [JWE] Added for SPH 1.0/CAP_SPH 0.9 fields that were not UserDef at that point but are now
	                            'APFDFactor'};
	                            %'PresumNumber';
	                            %'HwAzCounter'};    
                            
            % Tokens that only allow certain values
            tokens.valued = struct(...
                'Grid',        {{'Polar','Rect','Keystone','Unknown'}},...
       	        'Mode',        {{'Spotlight' ,'Stripmap','Circle','ArbitraryStripmap','SpotDwell','VideoCircle','VideoLine','UnknownMode7','GmtiWas','GmtiStt','HiPrfCc','SpotList','VideoMultiRes','MaritimeWas','Unknown'}},...
                'MoComp',      {{'SNL','TRDT'}},...
                'Geometry',    {{'Monostatic','Bistatic'}},...
                'GPSDatum',    {{'WGS-84'}},...
                'GPSDifCorOn', {{'Yes','No'}},...
                'GPSStatus',   {{'Ok', 'Gaps', 'Interpolated'}},...
                'GPSType',     {{'LLH','XYZ'}},...
                'SampleType',  {{'IQ','QI','MP','PM'}},...
                'ChirpType',   {{'Linear'}} );

        end
        
        %=================================================================%        
        % SPH data format is very flexible, however, this creates a problem for the
        % real time code.
        %
        % The CAP team has created a subset of the SPH file format which is
        % considered a valid CAP SPH data format. This file defines that tighter
        % specification so that our SPH writer will write "CAP Complient" SPH
        % files.
        % 
        % Reference: Data Format Document: CAP SPH Data format
        %            Revision 1.1 03/31/2011
        %=================================================================%        
        function [ cap ] = cap_sph_preamble_fields(radar_mode)
            %---------------------------------------------------------------------%
            % For the CAP SPH standard these fields must appear in order in the SPH
            % file:
            cap.fields = {
                'Version'
                'SystemSerialNum'
                'SoftwareVersionNum'
                'Mission'
                'Comment'
                'Grid'
                'Mode'
                'MoComp'
                'Geometry'
                'DateTime'
                'StartTime'
                'Sensor'
                'LightSpeed'
                'Npulses'
                'Nsamples'
                'Nchannels'
                'ChanDesc'
                'GPSDatum'
                'GPSDifCorOn'
                'GPSStatus'
                'GPSType'
                'SampleType'
                'DesCntrFreq'
                'DesElevAng'
                'DesRange'
                'DesGrndSqnt'
                'DesAzRes'
                'DesRngRes'
                'MagCal'
                'SigmaN'
                'ChirpType'
                'ADF'};

            %---------------------------------------------------------------------%
            %% Define Cap Structs

            % Note: These Cells were generated by passing an existing struct
            % through the function 'dumpstruct'. If the format changes there is no
            % reason to type all this out by hand, just dump structs from a new
            % valid CAP SPH data file that has been read in!
            
            %% Base Struct definitions
            RadarData = {
                'RxGain',                'float32',  4;
                'RxAtten',               'float32',  4;
                'SuggTxPower',           'float32',  4;
                'TxAtten',               'float32',  4;
                'TxPower',               'float32',  4;
                'TxPowerSource',         'int32',    4;
                'TxMaxDutyCycle',        'float32',  4;
                'MaxTxPower',            'float32',  4;
                'RangePaneFilterCoeff',  'float32',  44;
                'AzPreFilterCoeff',      'float32',  20;
                'HPFMeanSource',         'int32',    4;
                'IChanMean',             'float32',  4;
                'QChanMean',             'float32',  4;
                };

            AntennaData = {
                'AntCcwAzLimit',  'float32',  4;
                'AntCwAzLimit',   'float32',  4;
                'AntId',          'int32',    4;
                'AzBeamWidth',    'float64',  8;
                'AzPatCoeff',     'float64',  48;
                'ElBeamWidth',    'float64',  8;
                'ElevPatCoeff',   'float64',  72;
                'Gant',           'float32',  4;
                'GantTx',         'float32',  4;
                };
            
            RadarCoordinateFrame = {
                'x', 'float64', 8;
                'y', 'float64', 8;
                'z', 'float64', 8;
                };
                    
            GpsStatus = {
                'GpsReceiverUsed',         'int32',    4;
                'ReceiverKeyed',           'int32',    12;
                'DifferentialCorrection',  'int32',    12;
                'P1_std',                  'float64',  8;
                'P2_std',                  'float64',  8;
                'P3_std',                  'float64',  8;
                'V1_std',                  'float64',  8;
                'V2_std',                  'float64',  8;
                'V3_std',                  'float64',  8;
                };

            ConstStruct = {
                'CapSphVersion',                'float64',      8;
                'AntOverSamplingFactor',        'float32',      4;
                'APFDFactor',                   'int32',        4;
                'radarData',                    'RadarData',    112;
                'antennaData',                  'AntennaData',  160;
                'gpsStatus',                    'GpsStatus',    76;
                'DesChirpRate',                 'float64',      8;
                'DesValidPulseMotionInterval',  'uint32',       4;
                'PresumNumber',                 'int32',        4;
                'RFOA',                         'float64',      8;
                'ZeroPiBiasFlag',               'int32',        4;
                'GrazingForSampling',           'float64',      8;
                };

              ChanStruct = {
                'SwSyncWord',            'uint32',                4;
                'SwAzCounter',           'uint32',                4;
                'AzError',               'float32',               4;
                'ChirpRate',             'float64',               8;
                'ChirpRateDelta',        'float64',               8;
                'SwDebugReserved',        'uint32',               4;
                'RESERVED_1',            'int8',                  1;
                'RxPos',                 'SPHPos',                25;
                'RxTime',                'float64',               8;
                'RxTimeDelta',           'float64',               8;
                'StartF',                'float64',               8;
                'StartFDelta',           'float64',               8;
                'TiltError',             'float32',               4;
                'TRDelay',               'float64',               8;
                'TwistError',            'float32',               4;
                'VelDown',               'float32',               4;
                'VelEast',               'float32',               4;
                'VelNorth',              'float32',               4;
                'GimbalStopAz',          'float32',               4;
                'GimbalStopTilt',        'float32',               4;
                'GimbalStopTwist',       'float32',               4;
                'KeepOutViol',           'float32',               4;
                'radarCoordinateFrame',  'RadarCoordinateFrame',  24;
                'PhGroupAdcOverflows',   'uint32',                4;
                'AdcOverflowCount',      'float64',               8;
                'SRP',                   'SPHPos',                25;
                'AntPnt',                'SPHPos',                25;
                'PulseRepFreq',          'float32',               4;
                'NumPulsesInAir',        'uint32',                4;
                'SwSyncWord2',           'uint32',                4;
                'HwSyncWord',            'uint32',                4;
                'HwTimeCountUpper',      'uint32',                4;
                'HwTimeCountLower',      'uint32',                4;
                'HwAzCounter',           'uint32',                4;
                'HwStatusRegister',      'uint32',                4;
                'HwSyncWord2',           'uint32',                4;
                 % must be set exeternally based on radar parameters:
                'SampleData',            'cmplxs',                0;
                'HwChecksum',            'uint32',                4;
                'HwChecksumCheck',       'uint32',                4;
                };

            PulseStruct = {
                % must be set exeternally based on radar parameters:
                'ChannelData',    'ChanStruct',                   0;
            };
        
            %% MWAS specific modifications to the structures. They are mostly similar so easier to create base settings 
            % then modify that defintion to match MWAS requirements
            if strcmp('MaritimeWas', radar_mode)
                
                % Append new fields to RadarData and AntennaData structs
                RadarData   = [RadarData;   {'LossRF_dB',             'float32',  4}];
                AntennaData = [AntennaData; {'LossRadome_dB',         'float32',  4}];
                
                % Add new fields to const struct
                ConstStruct = [ConstStruct;                              
                              % PulseRepFreq moved from pulse header to const block for MWAS
                              {'PulseRepFreq',                 'float32',      4;}
                              % Non struct new MWAS specific fields
                              {'TxEffChirpRate',               'float64',      8;}
                              {'TxEffStartF',                  'float64',      8;}
                              {'TxEffStartPhase',              'float64',      8;}
                              {'TxEffDuration',                'float64',      8;}
                              {'ScanRate',                     'float64',      8;}
                              {'LossAtmosphere_dBpKm',         'float64',      8;}
                              {'GainNoiseFigure_dB'            'float64',      8;}
                              {'GainDRX_dBpW',                 'float64',      8;}];
                          
                % Replace PulseRepFreq with RESERVED_2
                idx = find(strcmp(ChanStruct(:,1),'PulseRepFreq'));
                ChanStruct(idx,:) = {'RESERVED_2', 'uint32', 4};
                
            end        

            %% Finalize definitions
            cap.struct_order = {'RadarCoordinateFrame',...
                                'RadarData',...
                                'AntennaData',...
                                'HwStatusRegister',...
                                'GpsStatus',...
                                'ConstStruct',...
                                'ChanStruct',...
                                'PulseStruct'};

            cap.datastructs.RadarCoordinateFrame = sph.cell2struct(RadarCoordinateFrame);
            cap.datastructs.RadarData            = sph.cell2struct(RadarData);
            cap.datastructs.AntennaData          = sph.cell2struct(AntennaData);
            cap.datastructs.GpsStatus            = sph.cell2struct(GpsStatus);
            cap.datastructs.ConstStruct          = sph.cell2struct(ConstStruct);
            cap.datastructs.ChanStruct           = sph.cell2struct(ChanStruct);
            cap.datastructs.PulseStruct          = sph.cell2struct(PulseStruct);

            HwStatusRegister = {
                'AuxDataValid', 'uint8', 0, 0;
                'BoardNumber',  'uint8', 2, 2;
                'RFC',          'uint8', 8, 8;
                'RFC2',         'uint8', 9, 9;
                'Channel',      'int32', 16, 23;
                };

            cap.BitDefs.HwStatusRegister = sph.cell2bitdef(HwStatusRegister);
        end
       
    %methods(Static, Access=protected)
        %=================================================================%        
        %READ_SPH_PREAMBLE populates a struct with information from a SPH file
        %
        % This function is typically called from READ_SPH_STREAM. Please see that
        % function for more information.
        %
        %  Date      Revision History
        %---------   -------------------------------------------------------------------
        % 06/29/09   rmn - Created
        % 11/03/11   rmn - Added support for UserDef / Arbitrary Structs
        % 02/07/12   rmn - Moved token definitions to external file
        %=================================================================%
        function preamble = read_sph_preamble(sph_fid)

            %
            tokens = sph.sph_tokens();

            % If we want to write out identical files we have to preserve the
            % ordering of structs and bitdefs - so we need to save the order
            preamble.struct_order = {};

            % read in the version number
            [tagString, remainder] = strtok(fgetl(sph_fid));
            if(~strncmp(tagString,'Version',7))
                msgId = 'sph:read_sph_preamble';
                msg = 'Version must be the first tag of preamble';
                throw(MException(msgId, msg));
            end
            expr='^\s*(?<major>\d+)\.(?<minor>\d+)\s*$';
            [found, preamble.Version]=regexp(remainder,expr,'start','names');

            % process other tags
            while(1)
                % Read a line
                lineText = fgetl(sph_fid);
                [tagString, firstRemainder] = strtok(lineText);

                switch(tagString)
                    case 'Comment'
                        if ~isfield(preamble,'Comment')
                            preamble.Comment = char();
                        end
                        preamble.Comment(end+1,:)=firstRemainder;

                    case 'DateTime'
                        expr='^\s*(?<year>\d{4})(?<month>\d{2})(?<day>\d{2})(?<hour>\d{2})(?<min>\d{2})(?<sec>\d{2})\s*$';
                        [found, DateTime]=regexp(firstRemainder,expr,'start','names');
                        fieldNames=fields(DateTime);
                        for i=1:length(fieldNames)
                            preamble.DateTime.(fieldNames{i})=str2double(DateTime.(fieldNames{i}));
                        end

                    case 'ChanDesc'
                        if(~isfield(preamble,'Nchannels'))
                           disp('WARNING: Nchannels must be assigned prior to any "ChanDesc" tags'); 
                        end
                        if(~isfield(preamble,'ChanDesc'))
                            preamble.ChanDesc = struct('number',{},'descriptor',{});
                        end
                        expr='^\s*(?<number>\d+)\s*(?<descriptor>\w+)\s*$';
                        [found, ChanDesc]=regexp(firstRemainder,expr,'start','names');
                        ChanDesc.number=str2double(ChanDesc.number);
                        preamble.ChanDesc(end+1)=ChanDesc;

                    case tokens.number
                         preamble.(tagString)=str2double(firstRemainder);

                    case tokens.word
                         %temp=textscan(firstRemainder,'%s',1);
                         %preamble.(tagString)=cell2mat(temp{1,:});
                         preamble.(tagString)=strtrim(firstRemainder);

                    case tokens.gpstype
                        temp = sscanf(firstRemainder,'%f %f %f');
                        preamble.(tagString).xat = temp(1);
                        preamble.(tagString).yon = temp(2);
                        preamble.(tagString).zae = temp(3);

                    case tokens.data_type
                        disp(tagString);

                    case fieldnames(tokens.valued)
                      try
                        temp = strtok(firstRemainder);
                        preamble.(tagString)=validatestring(temp,tokens.valued.(tagString));
                      catch
                         warning('%s NOT ONE OF EXPECTED VALUES FOR %s', temp, tagString); 
                      end

                    case 'EndPreamble'
                        break;

                    case 'BegBitDef'
                        [tmp, name] = bitdef_reader();
                        preamble.struct_order{end+1} = name;
                        preamble.BitDefs.(name)=tmp;

                    case {'BegChan','BegPulse','BegConst'}
                        name = strcat(tagString(4:end),'Struct');
                        preamble.struct_order{end+1} = name;
                        preamble.datastructs.(name)=struct_reader();

                    case {'BegStruct'}
                        [tmp, name] = struct_reader();
                        preamble.struct_order{end+1} = name;
                        preamble.datastructs.(name)=tmp;

                    otherwise
                        warning(['skipping.... ' tagString]);
                end
            end

            %Check for required tags

            if ~isfield(preamble,'BitDefs')
                preamble.BitDefs     = [];        
            end

            % helper functions
            %=============================================================%            
            function [bitStruct, name] = bitdef_reader()
                bitStruct = struct('name', {}, 'type', {}, 'begbit', {}, 'endbit', {});
                while(1)
                    lineText = fgetl(sph_fid);
                    [tagString, firstRemainder] = strtok(lineText);
                    switch(tagString)
                        case 'BitName'
                            name = textscan(firstRemainder, '%s');
                            name = char(name{1});

                        case 'EndBitDef'
                            break;

                        case 'Bits'
                            C = textscan(firstRemainder, '%s %s %n %n', 1);
                            bitStruct(end+1).name=char(C{1});
                            bitStruct(end).type=char(C{2});
                            bitStruct(end).begbit=C{3};
                            bitStruct(end).endbit=C{4};

                        otherwise
                            disp(['skipping.... ' tagString]);
                    end
                end
            end


            %=============================================================%
            function [dataStruct, name] = struct_reader()
                dataStruct = struct('name', {}, 'type', {}, 'bytes', {});
                while(1)
                    % Read a line
                    lineText = fgetl(sph_fid);
                    [tagString, firstRemainder] = strtok(lineText);
                    switch(tagString)
                        case {'UserDef','StructVar'}
                            %disp('UserDef/StructVar not yet handled');
                            C = textscan(firstRemainder, '%s %s %n', 1);
                            dataStruct(end+1).name=char(C{1});
                            dataStruct(end).type=char(C{2});
                            dataStruct(end).bytes=C{3}; 

                        case {'EndStruct','EndConst','EndChan','EndPulse'}
                            break;

                        case {'StructName'}
                            name = textscan(firstRemainder, '%s');
                            name = char(name{1});

                        case tokens.data_type
                            C = textscan(firstRemainder, '%s %n', 1);
                            tagName=tagString;
                            dataTypeName=char(C{1,1});
                            totalByteSize=C{1,2};
                            dataStruct(end+1).name=tagName;
                            dataStruct(end).type=dataTypeName;
                            dataStruct(end).bytes=totalByteSize;

                        otherwise
                            disp(['Assuming UserDef.... ' tagString]);
                            C = textscan(firstRemainder, '%s %n', 1);
                            dataStruct(end+1).name=tagString;
                            dataStruct(end).type=char(C{1});
                            dataStruct(end).bytes=C{2};

                    end
                end
            end
        end        

        %=================================================================%        
        %WRITE_SPH_PREAMBLE writes a SPH preamble to a file
        %
        % Given a preamble structure as read by READ_SPH_STRUCT, write this
        % preamble back to FID
        %
        %  Date      Revision History
        %---------   ------------------------------------------------------
        % 02/07/12   rmn - Created
        % 02/08/12   rmn - Added Option to force CAP SPH compliance
        %=================================================================%
        function write_sph_preamble(preamble, fid, varargin)

            if nargin < 3
                force_cap_compliance = 0;
            else
                force_cap_compliance = varargin{1};
            end

            tokens = sph.sph_tokens();
            tokens.valued = fieldnames(tokens.valued);

            % all known tokens
            tokens.known = [tokens.number; tokens.word; tokens.valued; tokens.data_type];

            if force_cap_compliance
                write_cap_sph();
            else
                write_generic_sph();
            end


            %---------------------------------------------------------------------%
            %---------------------------------------------------------------------%
            function write_cap_sph()

                % Set the correct sizes for the structs based on the size of our
                % sample data
                find_index = @(s, tag) (find(strcmp({s.name},tag),1,'first'));
                cap_fields = cap_sph_preamble_fields(preamble.Mode);
                cap_sd_index = find_index(cap_fields.datastructs.ChanStruct, 'SampleData');
                pre_sd_index = find_index(preamble.datastructs.ChanStruct,   'SampleData');
                cap_fields.datastructs.ChanStruct(cap_sd_index).bytes = preamble.datastructs.ChanStruct(pre_sd_index).bytes;
                cap_fields.datastructs.PulseStruct.bytes = sum([cap_fields.datastructs.ChanStruct.bytes]);

                % Check to make sure that we have the proper format
                if check_sph_cap_compatibility(preamble) ~= 0
                    fprintf(2, 'SPH Preamble not CAP Compliant\n');
                    fprintf(2, 'Please run ''check_sph_cap_compatibility()'' for more information,\n'); 
                    fprintf(2, 'and fix identified problems before saving SPH files\n');
                    return;
                end

                % If we made it this far we can go ahead and write out that data!
                %

                % First write out the preamble fields:
                fields = cap_fields.fields;
                for i=1:length(fields)
                    write_field(fields{i});
                end

                % Then the structs in order:
                write_ordered_struct_elements(cap_fields.datastructs, cap_fields.BitDefs, cap_fields.struct_order);

                fprintf(fid, 'EndPreamble\n');
            end

            %---------------------------------------------------------------------%
            function write_generic_sph()
                fields = fieldnames(preamble);
                for i=1:length(fields)
                    write_field(fields{i});
                end

                % Then the structs in order:
                write_ordered_struct_elements(preamble.datastructs, preamble.BitDefs, preamble.struct_order);

                fprintf(fid, 'EndPreamble\n');
            end

            %---------------------------------------------------------------------%
            function write_ordered_struct_elements(datastructs, bitdefs, order)
                for s = 1:length(order)
                    struct_name = order{s};
                    if isfield(datastructs, struct_name)
                        sph.pretty_print_struct(datastructs.(struct_name), struct_name, tokens.known, fid);
                    elseif isfield(bitdefs, struct_name)
                        sph.pretty_print_bitdef(bitdefs.(struct_name), struct_name, fid);
                    end
                end
            end

            %---------------------------------------------------------------------%
            function write_field(field)

                switch(field)

                    case 'Version'
                        fprintf(fid, 'Version     %s.%s\n', preamble.Version.major, preamble.Version.minor);

                    case 'SystemSerialNum' % just to get spaces right
                        fprintf(fid, 'SystemSerialNum     %d\n', preamble.SystemSerialNum);

                    case 'SoftwareVersionNum'
                        fprintf(fid, 'SoftwareVersionNum  %s\n', preamble.SoftwareVersionNum);
                        
                    case 'NumPulsesInAir'
                        fprintf(fid, 'NumPulsesInAir      %d\n', preamble.NumPulsesInAir);
			
                    case 'ChanDesc'
                        for c=1:length(preamble.ChanDesc)
                            fprintf(fid, '%-12s%d %s\n','ChanDesc', preamble.ChanDesc(c).number,preamble.ChanDesc(c).descriptor);
                        end

                    case 'DateTime'
                        fprintf(fid, '%-12s%04d%02d%02d%02d%02d%02d\n',...
                            'DateTime',...
                            preamble.DateTime.year,...
                            preamble.DateTime.month,...
                            preamble.DateTime.day,...
                            preamble.DateTime.hour,...
                            preamble.DateTime.min,...
                            preamble.DateTime.sec);

                    case 'SRP'
                        fprintf(fid, '%-12s%.7f  %.7f  %.7f\n',...
                            'SRP',...
                            preamble.SRP.xat,...
                            preamble.SRP.yon,...
                            preamble.SRP.zae);

                    case tokens.number
                        num = preamble.(field);
                        if (int16(num)==num)
                            format = '%-12s%d\n';
                        else
                            format = '%-12s%.9f\n';
                        end
                        fprintf(fid, format, field, num);

                    case ['Comment'; tokens.word; tokens.valued]
                        fprintf(fid, '%-12s%s\n', field, strtrim(preamble.(field)));

                    case {'datastructs','BitDefs','struct_order','dataFormat'}
                        % These aren't really fields in the SPH file preamble, they
                        % are added to the MATLAB struct

                    otherwise
                        disp(['ZZZ' field 'ZZZ']);

                end
            end

        end

        %---------------------------------------------------------------------%
        %---------------------------------------------------------------------%

        %---------------------------------------------------------------------%
        function pretty_print_bitdef(s, name, fid)
            fprintf(fid, 'BegBitDef\n');
            fprintf(fid, '  BitName  %s\n', name);

            field_names = {s.name};
            for i=1:length(field_names)
                fprintf(fid, '    Bits %-20s %-20s %d %d\n',s(i).name, s(i).type, s(i).begbit, s(i).endbit );
            end

            fprintf(fid, 'EndBitDef\n');
        end

        %---------------------------------------------------------------------%
        function pretty_print_struct(s, name, known_tokens, fid)

            switch name
                case {'ChanStruct','PulseStruct','ConstStruct'}
                    tag = name(1:length(name)-6);
                    fprintf(fid, ['Beg' tag '\n']);
                    print_internals(' ', sph.prefix_names(s, known_tokens), fid);
                    fprintf(fid, ['End' tag '\n']);

                otherwise
                    fprintf(fid, 'BegStruct\n');
                    fprintf(fid, '  StructName %s\n',name);
                    print_internals('     StructVar ', s, fid);
                    fprintf(fid, 'EndStruct\n');
            end

            %------%
            function print_internals(prefix, vars, fid)
                maxlength = @(x) max(cellfun(@length,x));
                pad = 2;
                name_length = maxlength({vars.name}) + pad + length('UserDef');
                desc_length = maxlength({vars.type}) + pad;
                format = ['%s %-' num2str(name_length)   's%-' num2str(desc_length) 's%d\n'];
                for i = 1:length(vars)       
                    fprintf(fid, format, prefix, vars(i).name, vars(i).type, vars(i).bytes);
                end
            end
        end

        %---------------------------------------------------------------------%
        % We need 'UserDef' infront of any variable that is not specificed by the
        % SPH documentation
        function vars = prefix_names(vars, known_tokens)
           for i=1:length(vars)
               if ~any(strcmp(vars(i).name, known_tokens)) || strcmp(vars(i).name, 'NumPulsesInAir')
                   vars(i).name = ['UserDef ' vars(i).name];
               end
           end
        end                        

        %=================================================================%
        function s = cell2struct(c)
            for r = 1:size(c,1)
               s(r) = struct('name', c{r,1},...
                             'type', c{r,2},...
                             'bytes', c{r,3});
            end
        end

        %=================================================================%
        function s = cell2bitdef(c)
            for r = 1:size(c,1)
               s(r) = struct('name', c{r,1},...
                             'type', c{r,2},...
                             'begbit', c{r,3},...
                             'endbit', c{r,4});
            end
        end

       
    end
    
end
